// Copyright (C) 2024  Umorpha Systems
// SPDX-License-Identifier: AGPL-3.0-or-later

package eclipse

import (
	"encoding"
	"fmt"
	"os"
	"path/filepath"
	"strings"
	"time"

	"sigs.k8s.io/yaml"
)

// Util ////////////////////////////////////////////////////////////////////////

func loadYAML[T any](filename string) (T, error) {
	bs, err := os.ReadFile(filename)
	if err != nil {
		var zero T
		return zero, err
	}

	var ret T
	if err := yaml.UnmarshalStrict(bs, &ret); err != nil {
		var zero T
		return zero, err
	}

	return ret, nil
}

type Duration struct {
	time.Duration
}

var _ encoding.TextUnmarshaler = (*Duration)(nil)

func (d *Duration) UnmarshalText(text []byte) error {
	var err error
	d.Duration, err = time.ParseDuration(string(text))
	return err
}

// Main config /////////////////////////////////////////////////////////////////

type Config struct {
	StateDir   string `json:"state_dir"`
	CacheDir   string `json:"cache_dir"`
	ActionsDir string `json:"actions_dir"`

	GitCacheFor       Duration `json:"cache_git_for"`
	ParallelDownloads int      `json:"parallel_downloads"`
}

func LoadConfig(filename string) (Config, error) {
	return loadYAML[Config](filename)
}

// Actions /////////////////////////////////////////////////////////////////////

type Action struct {
	Triggers []GitTrigger `json:"triggers"`
	Run      string       `json:"run"`
}

type GitTrigger struct {
	RepoURL string `json:"repo_url"`
	Rev     string `json:"rev"`
}

func (cfg Config) Actions() ([]Action, error) {
	entries, err := os.ReadDir(cfg.ActionsDir)
	if err != nil {
		return nil, err
	}
	var ret []Action
	for _, entry := range entries {
		if !strings.HasSuffix(entry.Name(), ".yml") && !strings.HasSuffix(entry.Name(), ".yaml") {
			continue
		}
		actions, err := loadYAML[[]Action](filepath.Join(cfg.ActionsDir, entry.Name()))
		if err != nil {
			return nil, fmt.Errorf("%q: %w", entry.Name(), err)
		}
		ret = append(ret, actions...)
	}
	return ret, nil
}
